using System;
using System.Collections.Generic;
using System.Text;
using System.Windows.Forms;
using System.Threading;
using System.Collections;

namespace LocalTimeReader
{
    public class LocalTimeReaderPort : ConnectedSerialPort
    {
        public enum PortStatus
        {
            NotConnected,
            ConnectionEstablished,
            ConnectionIdle,
            CheckConnected
        }

        private PortStatus status;
        private MainForm parent;

        private Queue queue1 = new Queue();
        private Queue sendMessageQueue;
        private Thread sendMessageThread;
        private ManualResetEvent sendEvent = new ManualResetEvent(false);

        private Queue queue2 = new Queue();
        private Queue receiveMessageQueue;
        private Thread receiveMessageThread;
        private ManualResetEvent receivedEvent = new ManualResetEvent(false);

        private System.Threading.Timer acknowledgeTimer = null;

        public LocalTimeReaderPort(MainForm prt)
        {
            parent = prt;

            sendMessageThread = new Thread(this.SendMessageThread);
            sendMessageQueue = Queue.Synchronized(queue1);
            sendMessageThread.Start();

            receiveMessageThread = new Thread(this.ReceiveMessageThread);
            receiveMessageQueue = Queue.Synchronized(queue2);
            receiveMessageThread.Start();

            acknowledgeTimer = new System.Threading.Timer(new TimerCallback(acknowledgeTimer_Elapsed), null, 5000, 5000);
            status = PortStatus.NotConnected;
        }

        private void acknowledgeTimer_Elapsed(object sender)
        {
            if (status == PortStatus.ConnectionEstablished)
                status =  PortStatus.ConnectionIdle;
            else if (status == PortStatus.ConnectionIdle)
            {
                status =  PortStatus.CheckConnected;
                SendAcknowledge();
            }
            else if (status == PortStatus.CheckConnected)
            {
                acknowledgeTimer.Change(Timeout.Infinite, Timeout.Infinite);
                MessageBox.Show("Communication lost - Timeout on acknowledge request",
                    "Communication Closed", MessageBoxButtons.OK, MessageBoxIcon.Information);

                this.ClosePort();
            }
        }

        protected override void ReceivedMessage(byte[] data, int length)
        {
            receiveMessageQueue.Enqueue(new Message(data, length));
            receivedEvent.Set();
        }

        private void ReceiveMessageThread()
        {
            while (true)
            {
                while (receiveMessageQueue.Count > 0)
                {
                    ReceivedMessage((Message)receiveMessageQueue.Dequeue());
                }

                receivedEvent.WaitOne();
                receivedEvent.Reset();
            }
        }

        private void ReceivedMessage(Message receivedMessage)
        {
            if (receivedMessage.Code == 0x03)
                status =  PortStatus.ConnectionEstablished;

            if (receivedMessage.Code == 0x01)
            {
                status = PortStatus.ConnectionEstablished;
//                SendAcknowledge();
                parent.ReceivedAcknowledgeRequest();
            }

            if (status ==  PortStatus.NotConnected)
                return;

            if (receivedMessage.Code == 0x05)
            {
                status = PortStatus.ConnectionEstablished;
                parent.ReceivedTime(receivedMessage.Index, receivedMessage.LocalTime, receivedMessage.RemoteTime);
            }
        }
        public void SendAcknowledge()
        {
            byte[] dataBytes = new byte[1];
            dataBytes[0] = 0x02;
            StartSendMessage(dataBytes);
        }

        public void SendTimeRequest(byte index, long time)
        {
            byte[] dataBytes = new byte[10];
            dataBytes[0] = 0x04;
            dataBytes[1] = index;
            (BitConverter.GetBytes(time)).CopyTo(dataBytes, 2);

            StartSendMessage(dataBytes);
        }

        private void StartSendMessage(byte[] data)
        {
            sendMessageQueue.Enqueue(data);
            sendEvent.Set();
        }

        private void SendMessageThread()
        {
            while (true)
            {
                while (sendMessageQueue.Count > 0)
                {
                    byte[] data = (byte[])sendMessageQueue.Dequeue();
                    SendMessage(data, data.Length);
                }
                sendEvent.WaitOne();
                sendEvent.Reset();

            }
        }

        public override void ClosePort()
        {
            receiveMessageThread.Abort();
            while (sendMessageQueue.Count != 0) ;
            Thread.Sleep(100);
            sendMessageThread.Abort();
            base.ClosePort();
            parent.PortClosing();
        }

    }
}
